﻿using GalaSoft.MvvmLight;
using Microsoft.Practices.ServiceLocation;
using Plane.CommunicationManagement;
using Plane.Util;
using Plane.Windows.IniHelper;
using Plane.Windows.Messages;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Media;
using System.Windows.Threading;

namespace Plane.FormationCreator.Formation
{
    public class RtcmManager: ObservableObject
    {
        private rtcm3 rtcm = new rtcm3();
        private Hashtable msgseen = new Hashtable();

        private ICommsSerial comPort;
        private CommModuleManager _commModuleManager = CommModuleManager.Instance;
        public ObservableCollection<RtcmInfo> rtcmInfoList { get; } = new ObservableCollection<RtcmInfo>();

        private bool _rtcmthreadrun = false;
        public bool Rtcmthreadrun
        {
            get { return _rtcmthreadrun; }
            set { Set(nameof(Rtcmthreadrun), ref _rtcmthreadrun, value); }
        }

        private int bps;
        private int bpsusefull;

        private int _bps = 0;
        public int Bps
        {
            get { return _bps; }
            set { Set(nameof(Bps), ref _bps, value); }
        }

        private int _bpsusefull = 0;
        public int Bpsusefull
        {
            get { return _bpsusefull; }
            set { Set(nameof(Bpsusefull), ref _bpsusefull, value); }
        }

        private string _messages_seen = "";
        public string Messages_seen
        {
            get { return _messages_seen; }
            set { Set(nameof(Messages_seen), ref _messages_seen, value); }
        }


        private DateTime baseTime = DateTime.Now;
        private DateTime gpsTime = DateTime.Now;
        private DateTime glonassTime = DateTime.Now;
        private DateTime beidouTime = DateTime.Now;

        private bool _baseState = false;
        public bool BaseState
        {
            get { return _baseState; }
            set { Set(nameof(BaseState), ref _baseState, value); }
        }

        private bool _gpsState = false;
        public bool GpsState
        {
            get { return _gpsState; }
            set { Set(nameof(GpsState), ref _gpsState, value); }
        }

        private bool _glonassState = false;
        public bool GlonassState
        {
            get { return _glonassState; }
            set { Set(nameof(GlonassState), ref _glonassState, value); }
        }

        private bool _beidouState = false;
        public bool BeidouState
        {
            get { return _beidouState; }
            set { Set(nameof(BeidouState), ref _beidouState, value); }
        }

        private double _stationLat;
        public double StationLat
        {
            get { return _stationLat; }
            set { Set(nameof(StationLat), ref _stationLat, value); }
        }

        private double _stationLng;
        public double StationLng
        {
            get { return _stationLng; }
            set { Set(nameof(StationLng), ref _stationLng, value); }
        }

        private double _stationAlt;
        public double StationAlt
        {
            get { return _stationAlt; }
            set { Set(nameof(StationAlt), ref _stationAlt, value); }
        }

        private string _stationTime;
        public string StationTime
        {
            get { return _stationTime; }
            set { Set(nameof(StationTime), ref _stationTime, value); }
        }

        private DispatcherTimer dispatcherTimer = null;
        public RtcmManager()
        {
            rtcm.ObsMessage += Rtcm_ObsMessage;
            dispatcherTimer = new DispatcherTimer();
            dispatcherTimer.Tick += new EventHandler(OnTimedEvent);
            dispatcherTimer.Interval = new TimeSpan(0, 0, 1);
            dispatcherTimer.Start();
        }

        private void OnTimedEvent(object sender, EventArgs e)
        {
            BaseState = baseTime > DateTime.Now;
            GpsState = gpsTime > DateTime.Now;
            GlonassState = glonassTime > DateTime.Now;
            BeidouState = beidouTime > DateTime.Now;

            Bps = bps;
            Bpsusefull = bpsusefull;
            bps = 0;
            bpsusefull = 0;

            StringBuilder sb = new StringBuilder();
            try
            {
                foreach (var item in msgseen.Keys)
                {
                    sb.Append(item + "=" + msgseen[item] + " ");
                }
            }
            catch
            {}
            Messages_seen = sb.ToString();
        }



        private void Rtcm_ObsMessage(object sender, EventArgs e)
        {
            List<rtcm3.ob> obs = sender as List<rtcm3.ob>;
            if (obs.Count == 0) return;
    
            List<RtcmInfo> partRtcmInfos;
            partRtcmInfos = rtcmInfoList.Where(o => o.Sys == obs[0].sys).ToList();
            while (partRtcmInfos.Count() < obs.Count)
            {
                RtcmInfo newInfo = new RtcmInfo();
                rtcmInfoList.Add(newInfo);
                partRtcmInfos.Add(newInfo);

            }

            while (partRtcmInfos.Count() > obs.Count)
            {
                RtcmInfo lastInfo = partRtcmInfos[partRtcmInfos.Count - 1];
                rtcmInfoList.Remove(lastInfo);
                partRtcmInfos.Remove(lastInfo);
            }

            if (partRtcmInfos.Count() == obs.Count)
            {
                for (int i = 0; i < obs.Count; i++)
                {
                    partRtcmInfos[i].Sys = obs[i].sys;
                    partRtcmInfos[i].Prn = obs[i].prn;
                    partRtcmInfos[i].Snr = obs[i].snr;
                }
            }
           
        }

        public string[] GetPortNames()
        {
            return SerialPort.GetPortNames();
        }

        public async Task Open(string CMB_serialport)
        {
            if (CMB_serialport == "") return;

            switch(CMB_serialport)
            {
               case "千寻":
                    //检测原点
                    FlightTaskManager _flightTaskManager = ServiceLocator.Current.GetInstance<FlightTaskManager>();
                    if (_flightTaskManager.OriginLat == 0 && _flightTaskManager.OriginLng == 0)
                    {
                        Alert.Show("作为参照的原点未设置，无法发送RTCM！", "提示");
                        return;
                    }

                    //千寻账号
                    string messagetips = "格式：http://差分账号:差分密码@域名:端口/坐标系\r\n例如：http://account:password@rtk.ntrip.qxwz.com:8002/RTCM32_GGB";
                    string url = "";
                    IniFiles inifilse = new IniFiles();
                    url = inifilse.IniReadvalue("Default", "RTKurl");
                    if (PlaneMessageBox.OnShow(messagetips, "发送RTK", ref url) == false) return;
                    inifilse.IniWritevalue("Default", "RTKurl", url);

                    //构造
                    comPort = new CommsNTRIP(url, _flightTaskManager.OriginLat, _flightTaskManager.OriginLng);
                    break;
                default:
                    comPort = new SerialPort();
                    comPort.PortName = CMB_serialport;
                    comPort.BaudRate = 57600;
                    break;
            }
            msgseen.Clear();
            comPort.Open();
            if (comPort.IsOpen)
            {
                Rtcmthreadrun = true;
                await RtcmLoop();
            }
        }

        public async Task Close()
        {
            await _commModuleManager.CloseRtcmLoop();
            Rtcmthreadrun = false;
            comPort.Close();
        }

        private async Task RtcmLoop()
        {
            int reconnecttimeout = 10;
            DateTime lastrecv = DateTime.MinValue;
            await _commModuleManager.StartRtcmLoop();
            while (Rtcmthreadrun)
            {
                try
                {
                    try
                    {
                        if ((DateTime.Now - lastrecv).TotalSeconds > reconnecttimeout || !comPort.IsOpen)
                        {
                            if (comPort is CommsNTRIP)
                            {
                            }
                            else
                            {
                                comPort.Close();
                                comPort.Open();
                            }
                            // reset timer
                            lastrecv = DateTime.Now;
                        }
                    }
                    catch { }
                
                    if (!comPort.IsOpen)
                    {
                        return;
                    }

                    //rtcm size=180
                    byte[] buffer = new byte[180];

                    int read = comPort.Read(buffer, 0, Math.Min(buffer.Length, comPort.BytesToRead));
                    if (read > 0)
                        lastrecv = DateTime.Now;

                    bps += read;
                    for (int a = 0; a < read; a++)
                    {
                        int seenmsg = -1;
                        // rtcm
                        if ((seenmsg = rtcm.Read(buffer[a])) > 0)
                        {
                            bpsusefull += rtcm.length;
                           
                            await _commModuleManager.InjectGpsRTCMDataAsync(rtcm.packet, rtcm.length);

                            string msgname = "Rtcm" + seenmsg;                          
                          
                            if (!msgseen.ContainsKey(msgname))
                                msgseen[msgname] = 0;
                            msgseen[msgname] = (int)msgseen[msgname] + 1;
                            
                            await ExtractBasePos(seenmsg);
                            //await seenRTCM(seenmsg);
                        }
                    }
                    await Task.Delay(10);
                }
                catch(Exception ex)
                {
                    Message.Show(ex.Message);
                }
            }
        }


        private async Task ExtractBasePos(int seen)
        {
            try
            {
                if (seen == 1005)
                {
                    var basepos = new rtcm3.type1005();
                    basepos.Read(rtcm.packet);
                    var pos = basepos.ecefposition;
                    double[] baseposllh = new double[3];
                    rtcm3.ecef2pos(pos, ref baseposllh);


                    StationLat = baseposllh[0] * rtcm3.R2D;
                    StationLng = baseposllh[1] * rtcm3.R2D;
                    StationAlt = baseposllh[2];
                    StationTime = DateTime.Now.ToString("HH:mm:ss");

                }
                else if (seen == 1006)
                {
                    var basepos = new rtcm3.type1006();
                    basepos.Read(rtcm.packet);
                    var pos = basepos.ecefposition;
                    double[] baseposllh = new double[3];
                    rtcm3.ecef2pos(pos, ref baseposllh);

                    StationLat = baseposllh[0] * rtcm3.R2D;
                    StationLng = baseposllh[1] * rtcm3.R2D;
                    StationAlt = baseposllh[2];
                    StationTime = DateTime.Now.ToString("HH:mm:ss");
                }

                switch (seen)
                {
                    case 1001:
                    case 1002:
                    case 1003:
                    case 1004:
                    case 1071:
                    case 1072:
                    case 1073:
                    case 1074:
                    case 1075:
                    case 1076:
                    case 1077:
                        gpsTime = DateTime.Now.AddSeconds(5);
                        break;
                    case 1005:
                    case 1006:
                    case 4072: // ublox moving base
                        baseTime = DateTime.Now.AddSeconds(20);
                        break;
                    case 1009:
                    case 1010:
                    case 1011:
                    case 1012:
                    case 1081:
                    case 1082:
                    case 1083:
                    case 1084:
                    case 1085:
                    case 1086:
                    case 1087:
                        glonassTime = DateTime.Now.AddSeconds(5);
                        break;
                    case 1121:
                    case 1122:
                    case 1123:
                    case 1124:
                    case 1125:
                    case 1126:
                    case 1127:
                        beidouTime = DateTime.Now.AddSeconds(5);
                        break;
                    default:
                        break;
                } 
            }
            catch (Exception ex)
            {
                Message.Show(ex.Message);
            }
            await Task.Delay(1);

        }
    }

    public class RtcmInfo: ObservableObject
    {
        private char sys;
        private byte prn;
        private byte snr;

        public RtcmInfo()
        {

        }

        public char Sys
        {
            get { return sys; }
            set { Set(nameof(Sys), ref sys, value); }
        }

        public byte Prn
        {
            get { return prn; }
            set { Set(nameof(Prn), ref prn, value); }
        }

        public byte Snr
        {
            get { return snr; }
            set { Set(nameof(Snr), ref snr, value); }
        }
    }
}
